@@ -7,5 +7,5 @@
  * file may be redistributed under the GNU Public License v2.
  */
 
-#define E2FSPROGS_VERSION "1.42.13"
-#define E2FSPROGS_DATE "17-May-2015"
+#define E2FSPROGS_VERSION "1.43.1"
+#define E2FSPROGS_DATE "08-Jun-2016"
